# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import copy
import logging
import time
from collections import OrderedDict
from contextlib import contextmanager
import six
from django.utils.functional import cached_property
from .client import get_es_client

logger = logging.getLogger('elasticindex')


class ElasticQuerySet(object):
    def __init__(self, model_cls, body=None, params=None):
        self.model_cls = model_cls
        self.body = body or {"query": {"match_all": {}}}
        self.params = params or {}
        self.latest_total_count = None
        self.latest_raw_result = None
        self.query_finished = False

    def __len__(self):
        return len(self.result_list)

    def __iter__(self):
        return iter(self.result_list)

    def __bool__(self):
        return bool(self.result_list)

    def __getitem__(self, k):
        """
        Retrieves an item or slice from the set of results.
        """
        if not isinstance(k, (slice,) + six.integer_types):
            raise TypeError
        assert ((not isinstance(k, slice) and (k >= 0)) or
                (isinstance(k, slice) and (k.start is None or k.start >= 0) and
                 (k.stop is None or k.stop >= 0))), \
            "Negative indexing is not supported."

        if self.query_finished:
            return self.result_list[k]

        if isinstance(k, slice):
            qs = self
            offset = 0
            if k.start is not None:
                offset = int(k.start)
                qs = qs.offset(offset)
            if k.stop is not None:
                limit = int(k.stop) - offset
                qs = qs.limit(limit)
            return list(qs)[::k.step] if k.step else qs

        qs = self.limit(1).offset(k)
        return list(qs)[0]

    def _clone(self):
        """
        :rtype: ElasticQuerySet
        """
        qs = self.__class__(
            self.model_cls, copy.deepcopy(self.body),
            copy.deepcopy(self.params))
        return qs

    @cached_property
    def result_list(self):
        self.query_finished = True
        return list(self.get_result())

    def get_result(self):
        """
        elasticsearch の search をそのまま実行
        :rtype: generator
        """
        with self.log_query():
            if self.params is not None:
                result = self.es_client.search(
                    index=self.model_cls.INDEX,
                    doc_type=self.model_cls.DOC_TYPE,
                    body=self.body, params=self.params)
            else:
                result = self.es_client.search(
                    index=self.model_cls.INDEX,
                    doc_type=self.model_cls.DOC_TYPE,
                    body=self.body)

        self.latest_total_count = result['hits']['total']
        self.latest_raw_result = result
        for hit in result['hits']['hits']:
            yield self.model_cls(hit)

    @cached_property
    def es_client(self):
        """
        :rtype: Elasticsearch
        """
        return get_es_client()

    def get_by_id(self, id):
        """
        Elasticsearch のIDで1件取得
        :param id:
        :return:
        """
        result = self.es_client.get(
            self.model_cls.INDEX, id, doc_type=self.model_cls.DOC_TYPE)
        self.latest_raw_result = result
        if not result['found']:
            raise self.model_cls.DoesNotExist(id)
        return self.model_cls(result)

    def delete_by_id(self, id):
        """
        Elasticsearch のIDで1件削除
        :param id: elasticsearch document id
        """
        result = self.es_client.delete(
            self.model_cls.INDEX, self.model_cls.DOC_TYPE, id)
        self.latest_raw_result = result
        return result

    def all(self):
        """
        :rtype: ElasticQuerySet
        """
        return self._clone()

    def limit(self, limit):
        """
        :rtype: ElasticQuerySet
        """
        o = self._clone()
        if limit is None:
            if 'size' in o.body:
                del o.body['size']
        else:
            o.body['size'] = limit
        return o

    def offset(self, offset):
        """
        :rtype: ElasticQuerySet
        """
        o = self._clone()
        if offset is None:
            if 'from' in o.body:
                del o.body['from']
        else:
            o.body['from'] = offset
        return o

    def query(self, filter_query_dict):
        """
        :param filter_query_dict:
          - {"match": {"product_id": 192}}
          - {"match_all": {}}  # default
          - {"multi_match": {
                "query": query_word,
                "fields": [
                    "upc", "title^3", "description", "authors",
                    "publishers", "tags", "keywords"]
            }}
          - {"bool": {
              "must": [
                {"match": {"is_used": True}},
                {"range": {"stock": {"gt": 0}}}
            ]}}

        :rtype: ElasticQuerySet
        """
        o = self._clone()
        o.body['query'] = filter_query_dict
        return o

    def set_body(self, body_dict):
        """
        replace query body
        """
        o = self._clone()
        o.body = body_dict
        return o

    def get(self, filter_query_dict):
        """
        1件取得
        複数件あってもエラーは出さず、黙って1件だけ返す
        """
        qs = self.query(filter_query_dict).limit(1)
        if not qs:
            raise self.model_cls.DoesNotExist(filter_query_dict)
        return qs[0]

    def count(self):
        """
        件数取得
        """
        if self.query_finished:
            return len(self.result_list)

        body = self.body.copy()
        if 'sort' in body:
            del body['sort']

        with self.log_query(label='count', body=body):
            if self.params is not None:
                result = self.es_client.count(
                    index=self.model_cls.INDEX,
                    doc_type=self.model_cls.DOC_TYPE,
                    body=body, params=self.params
                )
            else:
                result = self.es_client.count(
                    index=self.model_cls.INDEX,
                    doc_type=self.model_cls.DOC_TYPE,
                    body=body
                )
        self.latest_raw_result = result
        return result['count']

    def order_by(self, order_query_list):
        """
        sort パラメータをつける
        :type order_query_list: list, dict, string
        - "mz_score"
        - {"mz_score": "desc"}

        """
        o = self._clone()
        o.body['sort'] = order_query_list
        return o

    @property
    def log_query(self):
        """
        クエリをロギングするコンテクストマネージャ
        elasticsearch や elasticsearch.trace のロガーを
        DEBUG レベルで設定するともっと詳しく出る (結果が全部出る)
        """

        @contextmanager
        def _context(label='', body=None):
            start_time = time.time()
            yield
            elapsed_time = time.time() - start_time
            logger.debug('{}time:{}ms, body:{}'.format(
                '{}: '.format(label) if label else '',
                int(elapsed_time * 100), body or self.body))

        return _context

    def bulk(self, body):
        return self.es_client.bulk(
            body, index=self.model_cls.INDEX,
            doc_type=self.model_cls.DOC_TYPE)


class ElasticDocumentManager(object):
    """
    class ElasticDocumentManager(ElasticQuerySet)
    でもいいんだけど、インスタンス変数が汚れる可能性があるので
    クラスプロパティっぽい感じで、アクセスされるたびに新しいクエリセットを作ることにした
    """

    def __init__(self, model_cls, body=None, params=None):
        self.model_cls = model_cls

    def __get__(self, cls, owner):
        return ElasticQuerySet(self.model_cls)


class ElasticIndexManager(object):
    def __init__(self, model_cls):
        self.model_cls = model_cls

    @cached_property
    def mappings_properties(self):
        return OrderedDict(
            [
                (f_name, f.mapping)
                for f_name, f
                in self.model_cls._cached_fields().items()
                ])

    @cached_property
    def mappings(self):
        """
        インデックスの mappings の指定にそのまま使える dict
        """
        return {
            self.model_cls.DOC_TYPE: {
                "properties": self.mappings_properties
            }
        }

    def delete(self):
        """
        インデックスを削除
        :return:
        """
        es = get_es_client()
        es.indices.delete(self.model_cls.INDEX, ignore=[404, ])

    @cached_property
    def create_body_params(self):
        body = {"mappings": self.mappings}

        index_setting = getattr(self.model_cls, 'INDEX_SETTINGS', None)
        if index_setting:
            body["settings"] = index_setting
        return body

    def create(self):
        """
        インデックスを作成
        :return:
        """
        es = get_es_client()
        es.indices.create(
            self.model_cls.INDEX, self.create_body_params)

    def exists(self):
        """
        インデックスが存在するか
        """
        es = get_es_client()
        return es.indices.exists(self.model_cls.INDEX)


class ElasticDocumentMeta(type):
    def __new__(mcs, name, bases, attrs):
        c = super(ElasticDocumentMeta, mcs).__new__(
            mcs, name, bases, attrs)

        c.objects = ElasticDocumentManager(c)
        c.index = ElasticIndexManager(c)
        return c



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
import time
from django.test import TestCase
from .models import DummyModel, DummyESDocument, DummyESDocumentPresetIndex


class TestESDocumentTest(TestCase):
    def setUp(self):
        if DummyESDocument.index.exists():
            DummyESDocument.index.delete()
        DummyESDocument.index.create()
        # install fixtures.

        # Bulk update from model instances.
        DummyModel.objects.create(key='quick', value='brown fox')
        DummyModel.objects.create(key='jumps', value='over the')
        DummyESDocument.rebuild_index()

        # Update from single instance.
        d3 = DummyModel.objects.create(key='lazy', value='dogs.')
        DummyESDocument.rebuild_index_by_source_model(d3)

        # Update from dict
        DummyESDocument.update('id-manually', {'key': 'spam', 'value': 'eggs'})

        # Wait commit
        for i in range(10):
            try:
                result = DummyESDocument.objects.get(
                    {"term": {"value": "eggs"}})
                break
            except DummyESDocument.DoesNotExist:
                time.sleep(0.3)
                continue
        self.assertEqual(result.key, 'spam')

    def test_index_search(self):

        # Simple query
        results = DummyESDocument.objects.query({"match": {"key": "jumps"}})
        result = list(results)[0]
        self.assertEqual(result.value, 'over the')

        # OR query
        qs = DummyESDocument.objects.query(
            {"bool": {
                "should": [
                    {"match": {"value": "dogs"}},
                    {"match": {"value": "fox"}},

                ]}})
        qs = qs.order_by({"key": "desc"})
        result = qs[1]
        self.assertEqual(result.value, "dogs.")

    def tearDown(self):
        # teardown ES index
        DummyESDocument.index.delete()


class TestESDocumentPresetIndexTest(TestCase):
    def setUp(self):
        if DummyESDocumentPresetIndex.index.exists():
            DummyESDocumentPresetIndex.index.delete()
        DummyESDocumentPresetIndex.index.create()
        DummyESDocumentPresetIndex.update(
            'doc1', {
                'key': 'doc1',
                'text_k': "セキュリティのため、ローカルホストやローカルネットワークに"
                          "しかアクセスを許可していない Web アプリってあると思います。",
                'text_b': "例えば、オフィスで起動している社内サーバ。Jenkinsとか。"
                          "Wikiとか。サーバ監視ツールとか。例えば、本番環境で起動"
                          "している Docker コンテナの中で動いているWebツールとか。",
            })
        DummyESDocumentPresetIndex.update(
            'doc2', {
                'key': 'doc2',
                'text_k': "私の場合は、elasticsearch の head プラグインのWeb"
                          "管理画面を起動しているのですが、ローカルホストからしか"
                          "アクセスを許可してませんので、外からアクセスするには"
                          "一工夫必要です。",
                'text_b': "クライアントは Firefox に入ってますし、サーバは OpenSSH "
                          "に組み込まれていますので、別途ソフトウェアのインストールは"
                          "不要です。",
            })

        # Simple Query (and wait commit) (OMG)
        for i in range(10):
            try:
                result = DummyESDocumentPresetIndex.objects.get(
                    {"term": {"key": "doc2"}})
                break
            except DummyESDocumentPresetIndex.DoesNotExist:
                time.sleep(0.3)
                continue
        self.assertIn('不要です。', result.text_b)

    def test_index_kuromoji_1(self):
        results = DummyESDocumentPresetIndex.objects.query(
            {"match": {"text_k": "起動"}})
        r = list(results)
        self.assertEqual(len(r), 1)
        self.assertEqual(r[0].key, 'doc2')

    def test_index_kuromoji_2(self):
        results = DummyESDocumentPresetIndex.objects.query(
            {"match": {"text_k": "ネットワーク"}})
        r = list(results)
        self.assertEqual(len(r), 1)
        self.assertEqual(r[0].key, 'doc1')

    def test_index_bigram_1(self):
        results = DummyESDocumentPresetIndex.objects.query(
            {"match": {"text_b": "ソフトウ"}})
        r = list(results)
        self.assertEqual(len(r), 1)
        self.assertEqual(r[0].key, 'doc2')

    def test_index_bigram_2(self):
        results = DummyESDocumentPresetIndex.objects.query(
            {"match": {"text_b": "視ツ"}})
        r = list(results)
        self.assertEqual(len(r), 1)
        self.assertEqual(r[0].key, 'doc1')

    def test_index_bigram_3(self):
        results = DummyESDocumentPresetIndex.objects.query(
            {"match": {"text_b": "Firefoxサーバ"}})
        list(results)
        # r = list(results)
        # len(r) がここで0にならないといけない。が、なってない
        # 要 Elasticsearchの理解
        # self.assertEqual(len(r), 0)

    def tearDown(self):
        DummyESDocumentPresetIndex.index.delete()



# -*- coding: utf-8 -*-
"""
Elasticsearch を Django のモデルっぽく使うクラス
"""
from __future__ import unicode_literals
import logging
from collections import OrderedDict
from .client import get_es_client
from .managers import ElasticDocumentMeta
from .fields import ElasticDocumentField

logger = logging.getLogger('elasticindex')


class ElasticDocument(object, metaclass=ElasticDocumentMeta):  # flake8: NOQA
    """
    Elasticsearch のドキュメントを司るクラス

    本クラスの、.index は ElasticIndexManager のインスタンスが付く。
    ESのインデックス作成や削除などができる
    """

    INDEX = "default_index"
    DOC_TYPE = "default_doc_type"
    # インデックス生成時の settings辞書
    INDEX_SETTINGS = None

    source_model = None  # インデックス生成元モデル

    class DoesNotExist(Exception):
        pass

    class ResultKeyError(Exception):
        pass

    @classmethod
    def _fields(cls):
        """
        クラスに定義されている ElasticDocumentField を、再帰的に探す

        dir() + getattr() でも同じ処理ができるが、そうすると
        @property を暴発させてしまうので vars() + __bases__ で。

        :return: このクラスの持つ JsonMasterField の一覧
            キーがフィールド名(変数名), 値が JsonMasterField のインスタンス
        :rtype: dict
        """
        _fields_dict = OrderedDict()
        for base_class in cls.__bases__:
            if hasattr(base_class, '_fields'):
                _fields_dict.update(base_class._fields())

        for name, value in vars(cls).items():
            if isinstance(value, ElasticDocumentField):
                value.contribute_to_class(cls, name)
                _fields_dict[name] = value
        return _fields_dict

    @classmethod
    def _cached_fields(cls):
        """
        _fields の取得をクラス変数にキャッシュする

        比較的重い処理なので。
        """
        if not hasattr(cls, '_fields_cache'):
            cls._fields_cache = cls._fields()
        return cls._fields_cache

    @classmethod
    def rebuild_index(cls, limit=None, offset=None,
                      filtering_func=None, bulk_size=1000):
        """
        インデックスを再生成する
        :param limit:
        :param offset:
        :param filtering_func:
        :return:
        """
        client = get_es_client()
        qs = cls.source_model.objects.all()
        if filtering_func is not None:
            qs = filtering_func(qs)
        if offset:
            qs = qs[offset:offset + limit]
        elif limit:
            qs = qs[:limit]

        if not bulk_size:
            # non bulk mode
            logger.debug('No bulk mode.')
            for source_model in qs:
                logger.debug('source_model: {}'.format(source_model))
                client.index(
                    cls.INDEX, cls.DOC_TYPE,
                    cls.data_dict_for_index(source_model),
                    id=cls.get_id_of_source_model(source_model))
            return

        # bulk update
        def _get_bulk_body(qs):
            bulk_body = []
            for source_model in qs:
                logger.debug('source_model: {}'.format(source_model))
                bulk_body.append({'index': {
                    "_id": cls.get_id_of_source_model(source_model)}})
                bulk_body.append(cls.data_dict_for_index(source_model))
                if len(bulk_body) > bulk_size:
                    yield bulk_body
                    bulk_body = []
            if bulk_body:
                yield bulk_body

        for bulk_body in _get_bulk_body(qs):
            logger.debug('bulk updating.')
            # logger.debug('{}'.format(bulk_body))
            client.bulk(bulk_body, index=cls.INDEX, doc_type=cls.DOC_TYPE)

    @classmethod
    def update_bulk(cls, bulk_body):
        """
        バルク更新
        :type bulk_body: list
        """
        client = get_es_client()
        client.bulk(bulk_body, index=cls.INDEX, doc_type=cls.DOC_TYPE)

    @classmethod
    def update(cls, id, data_dict):
        """
        1レコードの更新 (insert/update)
        dict で直接内容を指定する
        通常はこれは使わず、rebuild_index もしくは rebuild_index_by_source_model を使う
        :type data_dict: dict
        """
        client = get_es_client()
        client.index(
            cls.INDEX, cls.DOC_TYPE,
            data_dict, id=id)

    @classmethod
    def rebuild_index_by_source_model(cls, source_model):
        """
        元モデルを1つ指定してインデックスを再生成
        :param source_model:
        :return:
        """
        cls.update(cls.get_id_of_source_model(source_model), cls.data_dict_for_index(source_model))

    @classmethod
    def data_dict_for_index(cls, source_model):
        params = {}
        for name, field in cls._cached_fields().items():
            params[name] = field.get_value_for_index_of_source_model(
                source_model)
        return params

    @classmethod
    def get_id_of_source_model(self, source_model):
        return source_model.pk

    def __init__(self, es_result):
        """
        ES検索結果からインスタンスを起こす
        """
        self.es_result = es_result
        self.es_id = es_result['_id']
        self.es_score = es_result['_score']
        es_source = es_result['_source']

        for field_name, field in self._cached_fields().items():
            if field_name not in es_source:
                raise self.ResultKeyError(field_name)
            value = field.get_value_from_index_source_value(
                es_source[field_name])
            setattr(self, field_name, value)



"""
settings に ELASTICINDEX_AWS_IAM があれば、
IAM クレデンシャルで Amazon ES への接続を行う
"""

import elasticsearch
from django.conf import settings


def get_es_client():
    """
    :return: Elasticsearch
    :rtype: Elasticsearch
    """
    if getattr(settings, 'ELASTICINDEX_AWS_IAM', None):
        return _get_es_client_aws()
    return elasticsearch.Elasticsearch(settings.ELASTICINDEX_HOSTS)


def _get_es_client_aws():
    """
    IAM を使って、Amazon ES にアクセスする場合
    :rtype: Elasticsearch
    """
    from requests_aws4auth import AWS4Auth
    iam = settings.ELASTICINDEX_AWS_IAM
    awsauth = AWS4Auth(
        iam['access_id'],
        iam['secret_key'],
        iam['region'],
        'es')

    return elasticsearch.Elasticsearch(
        hosts=settings.ELASTICINDEX_HOSTS,
        http_auth=awsauth,
        use_ssl=True,
        verify_certs=True,
        connection_class=elasticsearch.connection.RequestsHttpConnection
    )



class ElasticDocumentField(object):
    """
    ElasticDocumentようのField
    score = ElasticDocumentField(default=0) みたいに使う。

    :param default: 値を取得時、無かった場合と None だった場合のデフォルト値
    :param verbose_name: Django モデルフィールドの verbose_name の代用。
        指定はできるが内部では使っていないので、指定しても意味ない
    """

    class AttrNotFoundInSourceModel(Exception):
        pass

    def __init__(self, verbose_name=None, mapping=None,
                 source_attr_name=None, source_value_getter=None):
        self.verbose_name = verbose_name
        self.mapping = mapping or {}
        self.source_attr_name = source_attr_name
        self.source_value_getter = source_value_getter

    def contribute_to_class(self, cls, name):
        self.model = cls
        self.name = name

    def get_value_for_index_of_source_model(self, source_model):
        """
        Djangoモデルからインデックス用の値を取得
        """
        if self.source_value_getter:
            return self.source_value_getter(source_model)
        attr_name = self.source_attr_name or self.name
        if not hasattr(source_model, attr_name):
            raise self.AttrNotFoundInSourceModel(attr_name)
        return getattr(source_model, attr_name)

    def get_value_from_index_source_value(self, index_source_value):
        """
        Elasticsearch の、検索結果の _source の値からPython に使える値に変換
        ES から返ってくる値は、int や リストなどの方を保持したまま取得できるので
        通常は変換不要
        """
        return index_source_value



#!/usr/bin/env python
# coding: utf-8
from setuptools import setup
from elasticindex import __author__, __version__, __license__

setup(
    name='django-elasticindex',
    version=__version__,
    description='Shallow elasticsearch wrapper on Django',
    license=__license__,
    author=__author__,
    author_email='ytyng@live.jp',
    url='https://github.com/ytyng/elasticindex.git',
    keywords='Elasticsearch, Django, Python',
    packages=['elasticindex'],
    install_requires=['elasticsearch', 'requests_aws4auth'],
    entry_points={
    },
)



from django.db import models
from elasticindex.models import ElasticDocument, ElasticDocumentField as F


class DummyModel(models.Model):
    key = models.CharField(max_length=20, primary_key=True)
    value = models.TextField()


class DummyESDocument(ElasticDocument):
    INDEX = "elasticindex_test_index"
    DOC_TYPE = "elasticindex_test_doc"

    source_model = DummyModel

    key = F(mapping={"type": "string", "index": "not_analyzed"})
    value = F(mapping={"type": "string"})


class DummyESDocumentPresetIndex(ElasticDocument):
    INDEX = "elasticindex_test_index_preset_i"
    DOC_TYPE = "elasticindex_test_doc_preset_i"

    INDEX_SETTINGS = {
        "index": {
            "analysis": {
                "tokenizer": {
                    "kuromoji": {
                        "type": "kuromoji_tokenizer"
                    },
                    "bigram": {
                        "type": "nGram",
                        "min_gram": "2",
                        "max_gram": "2",
                        "token_chars": [
                            "letter",
                            "digit"
                        ]},
                },
                "analyzer": {
                    "kuromoji_analyzer": {
                        "type": "custom",
                        "tokenizer": "kuromoji_tokenizer"
                    },
                    "bigram_analyzer": {
                        "type": "custom",
                        "tokenizer": "bigram"
                    }
                }
            }
        },
    }

    key = F(mapping={"type": "string", "index": "not_analyzed"})
    text_k = F(mapping={
        "type": "string",
        "analyzer": "kuromoji_analyzer",
    })
    text_b = F(mapping={
        "type": "string",
        "tokenizer": "bigram",
        "analyzer": "bigram_analyzer",
    })
